@@ -15,6 +15,7 @@ package uyuni
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"net/http"
 	"net/url"
@@ -23,9 +24,7 @@ import (
 	"time"
 
 	"github.com/go-kit/log"
-	"github.com/go-kit/log/level"
 	"github.com/kolo/xmlrpc"
-	"github.com/pkg/errors"
 	"github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
 
@@ -47,13 +46,16 @@ const (
 	uyuniLabelProxyModule    = uyuniMetaLabelPrefix + "proxy_module"
 	uyuniLabelMetricsPath    = uyuniMetaLabelPrefix + "metrics_path"
 	uyuniLabelScheme         = uyuniMetaLabelPrefix + "scheme"
+
+	tokenDuration = 10 * time.Minute
 )
 
 // DefaultSDConfig is the default Uyuni SD configuration.
 var DefaultSDConfig = SDConfig{
-	Entitlement:     "monitoring_entitled",
-	Separator:       ",",
-	RefreshInterval: model.Duration(1 * time.Minute),
+	Entitlement:      "monitoring_entitled",
+	Separator:        ",",
+	RefreshInterval:  model.Duration(1 * time.Minute),
+	HTTPClientConfig: config.DefaultHTTPClientConfig,
 }
 
 func init() {
@@ -62,7 +64,7 @@ func init() {
 
 // SDConfig is the configuration for Uyuni based service discovery.
 type SDConfig struct {
-	Server           config.URL              `yaml:"server"`
+	Server           string                  `yaml:"server"`
 	Username         string                  `yaml:"username"`
 	Password         config.Secret           `yaml:"password"`
 	HTTPClientConfig config.HTTPClientConfig `yaml:",inline"`
@@ -96,14 +98,16 @@ type endpointInfo struct {
 // Discovery periodically performs Uyuni API requests. It implements the Discoverer interface.
 type Discovery struct {
 	*refresh.Discovery
-	apiURL       *url.URL
-	roundTripper http.RoundTripper
-	username     string
-	password     string
-	entitlement  string
-	separator    string
-	interval     time.Duration
-	logger       log.Logger
+	apiURL          *url.URL
+	roundTripper    http.RoundTripper
+	username        string
+	password        string
+	token           string
+	tokenExpiration time.Time
+	entitlement     string
+	separator       string
+	interval        time.Duration
+	logger          log.Logger
 }
 
 // Name returns the name of the Config.
@@ -114,22 +118,26 @@ func (c *SDConfig) NewDiscoverer(opts discovery.DiscovererOptions) (discovery.Di
 	return NewDiscovery(c, opts.Logger)
 }
 
+// SetDirectory joins any relative file paths with dir.
+func (c *SDConfig) SetDirectory(dir string) {
+	c.HTTPClientConfig.SetDirectory(dir)
+}
+
 // UnmarshalYAML implements the yaml.Unmarshaler interface.
 func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	*c = DefaultSDConfig
 	type plain SDConfig
 	err := unmarshal((*plain)(c))
-
 	if err != nil {
 		return err
 	}
-	if c.Server.URL == nil {
+	if c.Server == "" {
 		return errors.New("Uyuni SD configuration requires server host")
 	}
 
-	_, err = url.Parse(c.Server.String())
+	_, err = url.Parse(c.Server)
 	if err != nil {
-		return errors.Wrap(err, "Uyuni Server URL is not valid")
+		return fmt.Errorf("Uyuni Server URL is not valid: %w", err)
 	}
 
 	if c.Username == "" {
@@ -138,20 +146,16 @@ func (c *SDConfig) UnmarshalYAML(unmarshal func(interface{}) error) error {
 	if c.Password == "" {
 		return errors.New("Uyuni SD configuration requires a password")
 	}
-	return nil
+	return c.HTTPClientConfig.Validate()
 }
 
-func login(rpcclient *xmlrpc.Client, user string, pass string) (string, error) {
+func login(rpcclient *xmlrpc.Client, user, pass string, duration int) (string, error) {
 	var result string
-	err := rpcclient.Call("auth.login", []interface{}{user, pass}, &result)
+	err := rpcclient.Call("auth.login", []interface{}{user, pass, duration}, &result)
 	return result, err
 }
 
-func logout(rpcclient *xmlrpc.Client, token string) error {
-	return rpcclient.Call("auth.logout", token, nil)
-}
-
-func getSystemGroupsInfoOfMonitoredClients(rpcclient *xmlrpc.Client, token string, entitlement string) (map[int][]systemGroupID, error) {
+func getSystemGroupsInfoOfMonitoredClients(rpcclient *xmlrpc.Client, token, entitlement string) (map[int][]systemGroupID, error) {
 	var systemGroupsInfos []struct {
 		SystemID     int             `xmlrpc:"id"`
 		SystemGroups []systemGroupID `xmlrpc:"system_groups"`
@@ -200,8 +204,10 @@ func getEndpointInfoForSystems(
 
 // NewDiscovery returns a uyuni discovery for the given configuration.
 func NewDiscovery(conf *SDConfig, logger log.Logger) (*Discovery, error) {
-	var apiURL *url.URL
-	*apiURL = *conf.Server.URL
+	apiURL, err := url.Parse(conf.Server)
+	if err != nil {
+		return nil, err
+	}
 	apiURL.Path = path.Join(apiURL.Path, uyuniXMLRPCAPIPath)
 
 	rt, err := config.NewRoundTripperFromConfig(conf.HTTPClientConfig, "uyuni_sd")
@@ -234,7 +240,6 @@ func (d *Discovery) getEndpointLabels(
 	systemGroupIDs []systemGroupID,
 	networkInfo networkInfo,
 ) model.LabelSet {
-
 	var addr, scheme string
 	managedGroupNames := getSystemGroupNames(systemGroupIDs)
 	addr = fmt.Sprintf("%s:%d", networkInfo.Hostname, endpoint.Port)
@@ -271,15 +276,13 @@ func getSystemGroupNames(systemGroupsIDs []systemGroupID) []string {
 
 func (d *Discovery) getTargetsForSystems(
 	rpcClient *xmlrpc.Client,
-	token string,
 	entitlement string,
 ) ([]model.LabelSet, error) {
-
 	result := make([]model.LabelSet, 0)
 
-	systemGroupIDsBySystemID, err := getSystemGroupsInfoOfMonitoredClients(rpcClient, token, entitlement)
+	systemGroupIDsBySystemID, err := getSystemGroupsInfoOfMonitoredClients(rpcClient, d.token, entitlement)
 	if err != nil {
-		return nil, errors.Wrap(err, "unable to get the managed system groups information of monitored clients")
+		return nil, fmt.Errorf("unable to get the managed system groups information of monitored clients: %w", err)
 	}
 
 	systemIDs := make([]int, 0, len(systemGroupIDsBySystemID))
@@ -287,14 +290,14 @@ func (d *Discovery) getTargetsForSystems(
 		systemIDs = append(systemIDs, systemID)
 	}
 
-	endpointInfos, err := getEndpointInfoForSystems(rpcClient, token, systemIDs)
+	endpointInfos, err := getEndpointInfoForSystems(rpcClient, d.token, systemIDs)
 	if err != nil {
-		return nil, errors.Wrap(err, "unable to get endpoints information")
+		return nil, fmt.Errorf("unable to get endpoints information: %w", err)
 	}
 
-	networkInfoBySystemID, err := getNetworkInformationForSystems(rpcClient, token, systemIDs)
+	networkInfoBySystemID, err := getNetworkInformationForSystems(rpcClient, d.token, systemIDs)
 	if err != nil {
-		return nil, errors.Wrap(err, "unable to get the systems network information")
+		return nil, fmt.Errorf("unable to get the systems network information: %w", err)
 	}
 
 	for _, endpoint := range endpointInfos {
@@ -309,25 +312,27 @@ func (d *Discovery) getTargetsForSystems(
 	return result, nil
 }
 
-func (d *Discovery) refresh(ctx context.Context) ([]*targetgroup.Group, error) {
+func (d *Discovery) refresh(_ context.Context) ([]*targetgroup.Group, error) {
 	rpcClient, err := xmlrpc.NewClient(d.apiURL.String(), d.roundTripper)
 	if err != nil {
 		return nil, err
 	}
 	defer rpcClient.Close()
 
-	token, err := login(rpcClient, d.username, d.password)
-	if err != nil {
-		return nil, errors.Wrap(err, "unable to login to Uyuni API")
-	}
-	defer func() {
-		if err := logout(rpcClient, token); err != nil {
-			level.Debug(d.logger).Log("msg", "Failed to log out from Uyuni API", "err", err)
+	if time.Now().After(d.tokenExpiration) {
+		// Uyuni API takes duration in seconds.
+		d.token, err = login(rpcClient, d.username, d.password, int(tokenDuration.Seconds()))
+		if err != nil {
+			return nil, fmt.Errorf("unable to login to Uyuni API: %w", err)
 		}
-	}()
+		// Login again at half the token lifetime.
+		d.tokenExpiration = time.Now().Add(tokenDuration / 2)
+	}
 
-	targetsForSystems, err := d.getTargetsForSystems(rpcClient, token, d.entitlement)
+	targetsForSystems, err := d.getTargetsForSystems(rpcClient, d.entitlement)
 	if err != nil {
+		// Force the renewal of the token on next refresh.
+		d.tokenExpiration = time.Now()
 		return nil, err
 	}
 
