@@ -26,8 +26,10 @@ import (
 
 	"github.com/prometheus/common/model"
 
+	"github.com/prometheus/prometheus/model/histogram"
 	"github.com/prometheus/prometheus/model/labels"
 	"github.com/prometheus/prometheus/model/timestamp"
+	"github.com/prometheus/prometheus/promql/parser/posrange"
 	"github.com/prometheus/prometheus/util/strutil"
 )
 
@@ -53,7 +55,7 @@ type parser struct {
 
 	// Everytime an Item is lexed that could be the end
 	// of certain expressions its end position is stored here.
-	lastClosing Pos
+	lastClosing posrange.Pos
 
 	yyParser yyParserImpl
 
@@ -120,7 +122,7 @@ func (p *parser) Close() {
 
 // ParseErr wraps a parsing error with line and position context.
 type ParseErr struct {
-	PositionRange PositionRange
+	PositionRange posrange.PositionRange
 	Err           error
 	Query         string
 
@@ -129,27 +131,7 @@ type ParseErr struct {
 }
 
 func (e *ParseErr) Error() string {
-	pos := int(e.PositionRange.Start)
-	lastLineBreak := -1
-	line := e.LineOffset + 1
-
-	var positionStr string
-
-	if pos < 0 || pos > len(e.Query) {
-		positionStr = "invalid position:"
-	} else {
-
-		for i, c := range e.Query[:pos] {
-			if c == '\n' {
-				lastLineBreak = i
-				line++
-			}
-		}
-
-		col := pos - lastLineBreak
-		positionStr = fmt.Sprintf("%d:%d:", line, col)
-	}
-	return fmt.Sprintf("%s parse error: %s", positionStr, e.Err)
+	return fmt.Sprintf("%s: parse error: %s", e.PositionRange.StartPosInput(e.Query, e.LineOffset), e.Err)
 }
 
 type ParseErrors []ParseErr
@@ -168,6 +150,21 @@ func (errs ParseErrors) Error() string {
 	return "error contains no error message"
 }
 
+// EnrichParseError enriches a single or list of parse errors (used for unit tests and promtool).
+func EnrichParseError(err error, enrich func(parseErr *ParseErr)) {
+	var parseErr *ParseErr
+	if errors.As(err, &parseErr) {
+		enrich(parseErr)
+	}
+	var parseErrors ParseErrors
+	if errors.As(err, &parseErrors) {
+		for i, e := range parseErrors {
+			enrich(&e)
+			parseErrors[i] = e
+		}
+	}
+}
+
 // ParseExpr returns the expression parsed from the input.
 func ParseExpr(input string) (expr Expr, err error) {
 	p := NewParser(input)
@@ -214,14 +211,18 @@ func ParseMetricSelector(input string) (m []*labels.Matcher, err error) {
 
 // SequenceValue is an omittable value in a sequence of time series values.
 type SequenceValue struct {
-	Value   float64
-	Omitted bool
+	Value     float64
+	Omitted   bool
+	Histogram *histogram.FloatHistogram
 }
 
 func (v SequenceValue) String() string {
 	if v.Omitted {
 		return "_"
 	}
+	if v.Histogram != nil {
+		return v.Histogram.String()
+	}
 	return fmt.Sprintf("%f", v.Value)
 }
 
@@ -255,12 +256,12 @@ func ParseSeriesDesc(input string) (labels labels.Labels, values []SequenceValue
 }
 
 // addParseErrf formats the error and appends it to the list of parsing errors.
-func (p *parser) addParseErrf(positionRange PositionRange, format string, args ...interface{}) {
+func (p *parser) addParseErrf(positionRange posrange.PositionRange, format string, args ...interface{}) {
 	p.addParseErr(positionRange, fmt.Errorf(format, args...))
 }
 
 // addParseErr appends the provided error to the list of parsing errors.
-func (p *parser) addParseErr(positionRange PositionRange, err error) {
+func (p *parser) addParseErr(positionRange posrange.PositionRange, err error) {
 	perr := ParseErr{
 		PositionRange: positionRange,
 		Err:           err,
@@ -270,6 +271,10 @@ func (p *parser) addParseErr(positionRange PositionRange, err error) {
 	p.parseErrors = append(p.parseErrors, perr)
 }
 
+func (p *parser) addSemanticError(err error) {
+	p.addParseErr(p.yyParser.lval.item.PositionRange(), err)
+}
+
 // unexpected creates a parser error complaining about an unexpected lexer item.
 // The item that is presented as unexpected is always the last item produced
 // by the lexer.
@@ -342,9 +347,9 @@ func (p *parser) Lex(lval *yySymType) int {
 
 	switch typ {
 	case ERROR:
-		pos := PositionRange{
+		pos := posrange.PositionRange{
 			Start: p.lex.start,
-			End:   Pos(len(p.lex.input)),
+			End:   posrange.Pos(len(p.lex.input)),
 		}
 		p.addParseErr(pos, errors.New(p.yyParser.lval.item.Val))
 
@@ -354,7 +359,7 @@ func (p *parser) Lex(lval *yySymType) int {
 		lval.item.Typ = EOF
 		p.InjectItem(0)
 	case RIGHT_BRACE, RIGHT_PAREN, RIGHT_BRACKET, DURATION, NUMBER:
-		p.lastClosing = lval.item.Pos + Pos(len(lval.item.Val))
+		p.lastClosing = lval.item.Pos + posrange.Pos(len(lval.item.Val))
 	}
 
 	return int(typ)
@@ -412,7 +417,7 @@ func (p *parser) newAggregateExpr(op Item, modifier, args Node) (ret *AggregateE
 	ret = modifier.(*AggregateExpr)
 	arguments := args.(Expressions)
 
-	ret.PosRange = PositionRange{
+	ret.PosRange = posrange.PositionRange{
 		Start: op.Pos,
 		End:   p.lastClosing,
 	}
@@ -443,6 +448,147 @@ func (p *parser) newAggregateExpr(op Item, modifier, args Node) (ret *AggregateE
 	return ret
 }
 
+// newMap is used when building the FloatHistogram from a map.
+func (p *parser) newMap() (ret map[string]interface{}) {
+	return map[string]interface{}{}
+}
+
+// mergeMaps is used to combine maps as they're used to later build the Float histogram.
+// This will merge the right map into the left map.
+func (p *parser) mergeMaps(left, right *map[string]interface{}) (ret *map[string]interface{}) {
+	for key, value := range *right {
+		if _, ok := (*left)[key]; ok {
+			p.addParseErrf(posrange.PositionRange{}, "duplicate key \"%s\" in histogram", key)
+			continue
+		}
+		(*left)[key] = value
+	}
+	return left
+}
+
+func (p *parser) histogramsIncreaseSeries(base, inc *histogram.FloatHistogram, times uint64) ([]SequenceValue, error) {
+	return p.histogramsSeries(base, inc, times, func(a, b *histogram.FloatHistogram) *histogram.FloatHistogram {
+		return a.Add(b)
+	})
+}
+
+func (p *parser) histogramsDecreaseSeries(base, inc *histogram.FloatHistogram, times uint64) ([]SequenceValue, error) {
+	return p.histogramsSeries(base, inc, times, func(a, b *histogram.FloatHistogram) *histogram.FloatHistogram {
+		return a.Sub(b)
+	})
+}
+
+func (p *parser) histogramsSeries(base, inc *histogram.FloatHistogram, times uint64,
+	combine func(*histogram.FloatHistogram, *histogram.FloatHistogram) *histogram.FloatHistogram,
+) ([]SequenceValue, error) {
+	ret := make([]SequenceValue, times+1)
+	// Add an additional value (the base) for time 0, which we ignore in tests.
+	ret[0] = SequenceValue{Histogram: base}
+	cur := base
+	for i := uint64(1); i <= times; i++ {
+		if cur.Schema > inc.Schema {
+			return nil, fmt.Errorf("error combining histograms: cannot merge from schema %d to %d", inc.Schema, cur.Schema)
+		}
+
+		cur = combine(cur.Copy(), inc)
+		ret[i] = SequenceValue{Histogram: cur}
+	}
+
+	return ret, nil
+}
+
+// buildHistogramFromMap is used in the grammar to take then individual parts of the histogram and complete it.
+func (p *parser) buildHistogramFromMap(desc *map[string]interface{}) *histogram.FloatHistogram {
+	output := &histogram.FloatHistogram{}
+
+	val, ok := (*desc)["schema"]
+	if ok {
+		schema, ok := val.(int64)
+		if ok {
+			output.Schema = int32(schema)
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing schema number: %v", val)
+		}
+	}
+
+	val, ok = (*desc)["sum"]
+	if ok {
+		sum, ok := val.(float64)
+		if ok {
+			output.Sum = sum
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing sum number: %v", val)
+		}
+	}
+	val, ok = (*desc)["count"]
+	if ok {
+		count, ok := val.(float64)
+		if ok {
+			output.Count = count
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing count number: %v", val)
+		}
+	}
+
+	val, ok = (*desc)["z_bucket"]
+	if ok {
+		bucket, ok := val.(float64)
+		if ok {
+			output.ZeroCount = bucket
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing z_bucket number: %v", val)
+		}
+	}
+	val, ok = (*desc)["z_bucket_w"]
+	if ok {
+		bucketWidth, ok := val.(float64)
+		if ok {
+			output.ZeroThreshold = bucketWidth
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing z_bucket_w number: %v", val)
+		}
+	}
+
+	buckets, spans := p.buildHistogramBucketsAndSpans(desc, "buckets", "offset")
+	output.PositiveBuckets = buckets
+	output.PositiveSpans = spans
+
+	buckets, spans = p.buildHistogramBucketsAndSpans(desc, "n_buckets", "n_offset")
+	output.NegativeBuckets = buckets
+	output.NegativeSpans = spans
+
+	return output
+}
+
+func (p *parser) buildHistogramBucketsAndSpans(desc *map[string]interface{}, bucketsKey, offsetKey string,
+) (buckets []float64, spans []histogram.Span) {
+	bucketCount := 0
+	val, ok := (*desc)[bucketsKey]
+	if ok {
+		val, ok := val.([]float64)
+		if ok {
+			buckets = val
+			bucketCount = len(buckets)
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing %s float array: %v", bucketsKey, val)
+		}
+	}
+	offset := int32(0)
+	val, ok = (*desc)[offsetKey]
+	if ok {
+		val, ok := val.(int64)
+		if ok {
+			offset = int32(val)
+		} else {
+			p.addParseErrf(p.yyParser.lval.item.PositionRange(), "error parsing %s number: %v", offsetKey, val)
+		}
+	}
+	if bucketCount > 0 {
+		spans = []histogram.Span{{Offset: offset, Length: uint32(bucketCount)}}
+	}
+	return
+}
+
 // number parses a number.
 func (p *parser) number(val string) float64 {
 	n, err := strconv.ParseInt(val, 0, 64)
@@ -512,7 +658,7 @@ func (p *parser) checkAST(node Node) (typ ValueType) {
 
 		// opRange returns the PositionRange of the operator part of the BinaryExpr.
 		// This is made a function instead of a variable, so it is lazily evaluated on demand.
-		opRange := func() (r PositionRange) {
+		opRange := func() (r posrange.PositionRange) {
 			// Remove whitespace at the beginning and end of the range.
 			for r.Start = n.LHS.PositionRange().End; isSpace(rune(p.lex.input[r.Start])); r.Start++ { // nolint:revive
 			}
@@ -716,7 +862,7 @@ func (p *parser) newLabelMatcher(label, operator, value Item) *labels.Matcher {
 // addOffset is used to set the offset in the generated parser.
 func (p *parser) addOffset(e Node, offset time.Duration) {
 	var orgoffsetp *time.Duration
-	var endPosp *Pos
+	var endPosp *posrange.Pos
 
 	switch s := e.(type) {
 	case *VectorSelector:
@@ -756,7 +902,7 @@ func (p *parser) setTimestamp(e Node, ts float64) {
 		p.addParseErrf(e.PositionRange(), "timestamp out of bounds for @ modifier: %f", ts)
 	}
 	var timestampp **int64
-	var endPosp *Pos
+	var endPosp *posrange.Pos
 
 	timestampp, _, endPosp, ok := p.getAtModifierVars(e)
 	if !ok {
@@ -785,11 +931,11 @@ func (p *parser) setAtModifierPreprocessor(e Node, op Item) {
 	*endPosp = p.lastClosing
 }
 
-func (p *parser) getAtModifierVars(e Node) (**int64, *ItemType, *Pos, bool) {
+func (p *parser) getAtModifierVars(e Node) (**int64, *ItemType, *posrange.Pos, bool) {
 	var (
 		timestampp **int64
 		preprocp   *ItemType
-		endPosp    *Pos
+		endPosp    *posrange.Pos
 	)
 	switch s := e.(type) {
 	case *VectorSelector:
